/*
 *	plugin_caf.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>
#include <sp/spOutputPlugin.h>

#define SP_CAF_TYPE_LPCM_BE 0
#define SP_CAF_TYPE_LPCM_LE 1
#define SP_CAF_TYPE_ULAW 2
#define SP_CAF_TYPE_ALAW 3
#define SP_NUM_CAF_TYPE 4
#if BYTE_ORDER == LITTLE_ENDIAN
#define SP_CAF_TYPE_DEFAULT SP_CAF_TYPE_LPCM_LE
#else
#define SP_CAF_TYPE_DEFAULT SP_CAF_TYPE_LPCM_BE
#endif

static const char *sp_caf_file_type_list[] = {
    "CAF",
    "CAF",
    "CAF",
    "CAF",
    NULL,
};
static const char *sp_caf_file_desc_list[] = {
    "Core Audio Format (Linear PCM, Big Endian)",
    "Core Audio Format (Linear PCM, Little Endian)",
    "Core Audio Format (mu-law)",
    "Core Audio Format (A-law)",
    NULL,
};
static const char *sp_caf_file_filter_list[] = {
    "*.caf",
    "*.caf",
    "*.caf",
    "*.caf",
    NULL,
};
    
typedef struct _spCafPluginInstance
{
    int type_index;
    
    FILE *fp;
    spLong64 current_pos;
    spLong64 total_duration;
    spSongInfoV2 song_info;

    int swap;
    int samp_bit;
    
    double sampleRate;
    long channelsPerFrame;
    unsigned long bitsPerChannel;

    spCafHeader cafHeader;
} *spCafPluginInstance;

static spBool spInitPluginCaf(const char *lang)
{
    /*spSetDebugLevel(10);*/

    return SP_TRUE;
}

static spBool spFreePluginCaf(void)
{
    return SP_TRUE;
}

static void *spInitPluginInstanceCaf(const char *lang)
{
    spCafPluginInstance instance;
    
    instance = xalloc(1, struct _spCafPluginInstance);
    memset(instance, 0, sizeof(struct _spCafPluginInstance));
    instance->type_index = SP_CAF_TYPE_DEFAULT;
    
    instance->fp = NULL;
    instance->current_pos = 0;
    instance->total_duration = 0;

    spInitSongInfoV2(&instance->song_info);
    
    instance->sampleRate = 44100.0;
    instance->channelsPerFrame = 2;
    instance->bitsPerChannel = 16;
    
    spDebug(50, "spInitPluginInstanceCaf", "done: instance = %ld\n", (long)instance);
    
    return (void *)instance;
}

static spBool spFreePluginInstanceCaf(void *instance)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;

    if (pinstance != NULL) {
	xfree(pinstance);
    }
    
    return SP_TRUE;
}

static spBool spSetPluginFileTypeCaf(void *instance, int index)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    pinstance->type_index = index;
    
    return SP_TRUE;
}

static int spGetPluginFileTypeCaf(void *instance)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    return pinstance->type_index;
}

static spBool spSetPluginSongInfoCaf(void *instance, spSongInfo *song_info)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    spCopySongInfoV2(&pinstance->song_info, (spSongInfoV2 *)song_info);

    return SP_TRUE;
}

static spBool spGetPluginSongInfoCaf(void *instance, spSongInfo *song_info)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    spCopySongInfoV2((spSongInfoV2 *)song_info, &pinstance->song_info);

    return SP_TRUE;
}

static const char *spGetPluginBestSuffixCaf(void *instance)
{
    return ".caf";
}

spBool spIsSampleBitSupportedCaf(int samp_bit)
{
    if (samp_bit == 8 || samp_bit == 16 || samp_bit == 24
	|| samp_bit == 32 || samp_bit == 33 || samp_bit == 64) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spBool spSetPluginSampleBitCaf(void *instance, int samp_bit)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;

    spDebug(50, "spSetPluginSampleBitCaf", "samp_bit = %d\n", samp_bit);

    if (spIsSampleBitSupportedCaf(samp_bit) == SP_TRUE) {
	pinstance->samp_bit = samp_bit;
	if (samp_bit == 33) {
	    pinstance->bitsPerChannel = 32;
	} else {
	    pinstance->bitsPerChannel = samp_bit;
	}
    
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spBool spGetPluginSampleBitCaf(void *instance, int *samp_bit)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;

    *samp_bit = pinstance->samp_bit;
    
    return SP_TRUE;
}

static spBool spSetPluginChannelCaf(void *instance, int num_channel)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    pinstance->channelsPerFrame = num_channel;
    spDebug(50, "spSetPluginChannelCaf", "mChannelsPerFrame = %ld\n", pinstance->channelsPerFrame);
    
    return SP_TRUE;
}

static spBool spGetPluginChannelCaf(void *instance, int *num_channel)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    *num_channel = pinstance->channelsPerFrame;
    
    return SP_TRUE;
}

static spBool spSetPluginSampleRateCaf(void *instance, double samp_rate)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    pinstance->sampleRate = samp_rate;
    spDebug(50, "spSetPluginSampleRateCaf", "mSampleRate = %f\n", pinstance->sampleRate);
    
    return SP_TRUE;
}

static spBool spGetPluginSampleRateCaf(void *instance, double *samp_rate)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    *samp_rate = pinstance->sampleRate;
    
    return SP_TRUE;
}

static spBool spSetPluginOtherInfoCaf(void *instance, const char *id, char *data)
{
#if 0
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
#endif
    
    return SP_FALSE;
}

static spBool spGetPluginOtherInfoCaf(void *instance, const char *id, char *data)
{
#if 0
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
#endif
    
    return SP_FALSE;
}
