/*
 *	input_monkey.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spInputPluginP.h>
#include <sp/spPluginMain.h>

#include "plugin_monkey.cpp"

static const char *sp_monkey_file_type_list[] = {
    "MONKEY",
    NULL,
};
static const char *sp_monkey_file_desc_list[] = {
    "Monkey's Audio",
    NULL,
};
static const char *sp_monkey_file_filter_list[] = {
    "*.ape",
    NULL,
};
    
static spBool spIsSupportedByPluginMonkey(const char *filename);

static spPluginError spOpenPluginMonkey(void *instance, const char *filename, const char *mode);
static spBool spClosePluginMonkey(void *instance);
static spPluginState spGetPluginStateMonkey(void *instance);
static spLong spGetPluginCurrentPositionMonkey(void *instance);
static spBool spStopPluginMonkey(void *instance);
static spBool spPausePluginMonkey(void *instance);
static spBool spRestartPluginMonkey(void *instance);

static long spReadPluginMonkey(void *instance, char *data, long length);
static spBool spSeekPluginMonkey(void *instance, spLong pos);
static spLong spGetPluginTotalLengthMonkey(void *instance);


static spInputPluginRec sp_input_plugin_monkey = {
    NULL,
    NULL,

    SP_PLUGIN_INPUT,
    "MONKEY",
    3,
    SP_PLUGIN_PRIORITY_HIGHER,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginMonkey,
    spFreePluginMonkey,
    "Monkey's Audio Input Plugin",
    "Monkey's Audio Input Plugin Version 0.3\nCopyright(C) Hideki Banno\nBased on Monkey's Audio SDK (http://www.monkeysaudio.com/developers.html)\nMonkey's Audio website: http://www.monkeysaudio.com/",
    
    spInitPluginInstanceMonkey,
    spFreePluginInstanceMonkey,
    NULL,
    NULL,

    SP_PLUGIN_DEVICE_FILE,
    sp_monkey_file_type_list,
    sp_monkey_file_desc_list,
    sp_monkey_file_filter_list,
    spIsSupportedByPluginMonkey,
    spSetPluginFileTypeMonkey,  
    spGetPluginFileTypeMonkey,
    (SP_SONG_TITLE_MASK | SP_SONG_ARTIST_MASK | SP_SONG_COMMENT_MASK
     | SP_SONG_GENRE_MASK | SP_SONG_RELEASE_MASK | SP_SONG_COPYRIGHT_MASK
     | SP_SONG_TRACK_MASK | SP_SONG_ALBUM_MASK),
    spSetPluginSongInfoMonkey,
    spGetPluginSongInfoMonkey,
    spGetPluginBestSuffixMonkey,
    
    NULL,  
    NULL,
    NULL,
    NULL,  
    NULL,
    NULL,
    NULL,

    spSetPluginSampleBitMonkey,   
    spGetPluginSampleBitMonkey,   
    spSetPluginChannelMonkey,     
    spGetPluginChannelMonkey,     
    spSetPluginSampleRateMonkey,  
    spGetPluginSampleRateMonkey,  
    spSetPluginOtherInfoMonkey,   
    spGetPluginOtherInfoMonkey,
    
    spOpenPluginMonkey,              
    spClosePluginMonkey,             
    spGetPluginStateMonkey,          
    spGetPluginCurrentPositionMonkey,
    spStopPluginMonkey,              
    spPausePluginMonkey,             
    spRestartPluginMonkey,

    spReadPluginMonkey,             
    spSeekPluginMonkey,
    spGetPluginTotalLengthMonkey,
};

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_input_plugin_monkey;
}

static spBool spIsSupportedByPluginMonkey(const char *filename)
{
    int nRetVal = 0;
    APEChar filename2[SP_MAX_PATHNAME];

    convertStringToAPEString(filename, filename2, SP_MAX_PATHNAME);
    spDebug(10, "spIsSupportedByPluginMonkey", "filename = %s, filename2 = %ls\n", filename, filename2);

    IAPEDecompress * pAPEDecompress = CreateIAPEDecompress(filename2, &nRetVal);
    if (pAPEDecompress == NULL)
    {
	spDebug(1, "spIsSupportedByPluginMonkey", "Error opening APE file (error code: %d)\r\n", nRetVal);
	return SP_FALSE;
    }

    delete pAPEDecompress;

    return SP_TRUE;
}

static void getFieldValue(spConverter conv, CAPETagField *pTagField, char *value)
{
    char *p;
    char buf[SP_SONG_INFO_SIZE];

    spStrCopy(buf, MIN(pTagField->GetFieldValueSize() + 1, SP_SONG_INFO_SIZE), 
	      (char *)pTagField->GetFieldValue());

    if (conv != NULL) {
	if ((p = xspConvert(conv, buf)) != NULL) {
	    strcpy(value, p);
	    spDebug(10, "getFieldValue", "value = %s\n", value);
	    xfree(p);
	}
    } else {
	strcpy(value, buf);
    }

    return;
}

#if MAC_VERSION_NUMBER >= 3980
#define fieldeq(s1, s2) ((s1 != NULL) && (s2 != NULL) && (_wcsicmp((const wchar_t *)(s1), (const wchar_t *)(s2)) == 0) ? 1 : 0)
#else
#define fieldeq(s1, s2) ((s1 != NULL) && (s2 != NULL) && (strcmp((const char *)(s1), (const char *)(s2)) == 0) ? 1 : 0)
#endif

static spBool getComments(spMonkeyPluginInstance pinstance, spSongInfo *song_info)
{
    int i;
    spBool flag = SP_FALSE;
    CAPETag *pAPETag;

    spDebug(80, "getComments", "in\n");
    
    pAPETag = (CAPETag *)pinstance->pAPEDecompress->GetInfo(APE_INFO_TAG);
    spDebug(80, "getComments", "pAPETag = %lx\n", (unsigned long)pAPETag);
    spDebug(80, "getComments", "pAPETag->GetHasAPETag() = %d\n", pAPETag->GetHasAPETag());
    spDebug(80, "getComments", "pAPETag->GetHasID3Tag() = %d\n", pAPETag->GetHasID3Tag());
    
    if (pAPETag->GetHasID3Tag() || pAPETag->GetHasAPETag()) {
	CAPETagField *pTagField;
	spConverter conv = NULL;
#if MAC_VERSION_NUMBER >= 3980
	const APEChar *field_name;
#else
	BOOL bFirst = TRUE;
	const char *field_name;
#endif

	if (CURRENT_APE_TAG_VERSION > 1000) {
	    /* tag version 2+ uses UTF8 */
            spDebug(80, "getComments", "call spOpenConverter\n");
	    conv = spOpenConverter("UTF-8", NULL);
	}

	for (i = 0;; i++)
	{
            spDebug(80, "getComments", "i = %d\n", i);
#if MAC_VERSION_NUMBER >= 3980
	    if ((pTagField = pAPETag->GetTagField(i)) == NULL) {
		break;
	    }
#else
	    if (pAPETag->GetNextTagField(bFirst, &pTagField) == FALSE) {
		break;
	    }
	    bFirst = FALSE;
#endif

	    field_name = pTagField->GetFieldName(); 

	    if (fieldeq(field_name, APE_TAG_FIELD_TITLE)) {
		getFieldValue(conv, pTagField, song_info->title);
		song_info->info_mask |= SP_SONG_TITLE_MASK;
	    } else if (fieldeq(field_name, APE_TAG_FIELD_ARTIST)) {
		getFieldValue(conv, pTagField, song_info->artist);
		song_info->info_mask |= SP_SONG_ARTIST_MASK;
	    } else if (fieldeq(field_name, APE_TAG_FIELD_ALBUM)) {
		getFieldValue(conv, pTagField, song_info->album);
		song_info->info_mask |= SP_SONG_ALBUM_MASK;
	    } else if (fieldeq(field_name, APE_TAG_FIELD_COMMENT)) {
		getFieldValue(conv, pTagField, song_info->comment);
		song_info->info_mask |= SP_SONG_COMMENT_MASK;
	    } else if (fieldeq(field_name, APE_TAG_FIELD_YEAR)) {
		getFieldValue(conv, pTagField, song_info->release);
		song_info->info_mask |= SP_SONG_RELEASE_MASK;
	    } else if (fieldeq(field_name, APE_TAG_FIELD_TRACK)) {
		song_info->track = atoi(pTagField->GetFieldValue());
		song_info->info_mask |= SP_SONG_TRACK_MASK;
	    } else if (fieldeq(field_name, APE_TAG_FIELD_GENRE)) {
		getFieldValue(conv, pTagField, song_info->genre);
		song_info->info_mask |= SP_SONG_GENRE_MASK;
	    } else if (fieldeq(field_name, APE_TAG_FIELD_COPYRIGHT)) {
		getFieldValue(conv, pTagField, song_info->copyright);
		song_info->info_mask |= SP_SONG_COPYRIGHT_MASK;
	    } else {
		/* not supported */
	    }
	}
	flag = SP_TRUE;

	if (conv != NULL) {
	    spCloseConverter(conv);
	}
    }


    return flag;
}

static spPluginError spOpenPluginMonkey(void *instance, const char *filename, const char *mode)
{
    int nRetVal = 0;
    APEChar filename2[SP_MAX_PATHNAME];
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;

    spDebug(80, "spOpenPluginMonkey", "in\n");
    
    spInitSongInfo(&pinstance->song_info);
    
    if (strnone(filename)) {
	return SP_PLUGIN_ERROR_OPEN;
    }
    convertStringToAPEString(filename, filename2, SP_MAX_PATHNAME);
    spDebug(80, "spOpenPluginMonkey", "convertStringToAPEString done\n");
    
    /* open file */
    if ((pinstance->pAPEDecompress = CreateIAPEDecompress(filename2, &nRetVal)) == NULL) {
	spDebug(1, "spOpenPluginMonkey", "Error opening APE file (error code: %d)\r\n", nRetVal);
	return SP_PLUGIN_ERROR_OPEN;
    }

    spDebug(10, "spOpenPluginMonkey", "filename = %s\n", filename);

    pinstance->compression_level = pinstance->pAPEDecompress->GetInfo(APE_INFO_COMPRESSION_LEVEL);
    pinstance->num_channel = pinstance->pAPEDecompress->GetInfo(APE_INFO_CHANNELS);
    pinstance->samp_rate = pinstance->pAPEDecompress->GetInfo(APE_INFO_SAMPLE_RATE);
    pinstance->samp_bit = pinstance->pAPEDecompress->GetInfo(APE_INFO_BITS_PER_SAMPLE);

    pinstance->total_length = pinstance->pAPEDecompress->GetInfo(APE_DECOMPRESS_TOTAL_BLOCKS);

    pinstance->current_pos = 0;

    getComments(pinstance, &pinstance->song_info);

    spDebug(80, "spOpenPluginMonkey", "done\n");
    
    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spClosePluginMonkey(void *instance)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;
    
    if (pinstance->pAPEDecompress != NULL) {
	delete pinstance->pAPEDecompress;
	
	pinstance->pAPEDecompress = NULL;
	pinstance->current_pos = 0;
    }

    return SP_TRUE;
}

static spPluginState spGetPluginStateMonkey(void *instance)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;
    
    if (pinstance->current_pos > 0) {
	return SP_PLUGIN_STATE_START;
    }

    return SP_PLUGIN_STATE_STOP;
}

static spLong spGetPluginCurrentPositionMonkey(void *instance)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;
    
    return pinstance->current_pos;
}

static spBool spStopPluginMonkey(void *instance)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;

    if (pinstance->pAPEDecompress == NULL) return SP_FALSE;

    if (pinstance->pAPEDecompress->Seek(0) != 0) { 
	return SP_FALSE; 
    }
    pinstance->current_pos = 0;

    return SP_TRUE;
}

static spBool spPausePluginMonkey(void *instance)
{

    return SP_TRUE;
}

static spBool spRestartPluginMonkey(void *instance)
{

    return SP_TRUE;
}

static long spReadPluginMonkey(void *instance, char *data, long length)
{
    spLong pos;
    long nread;
    int nRetVal = 0;
    intn nBlocksRetrieved;
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;    
    
    if (pinstance->pAPEDecompress == NULL) return -1;

    nRetVal = pinstance->pAPEDecompress->GetData(data, (intn)(length / pinstance->num_channel), &nBlocksRetrieved);
    if (nRetVal != ERROR_SUCCESS) {
	spDebug(1, "spReadPluginMonkey", "Decoding error (%d)\r\n", nRetVal);
	return -1;
    }

    nread = nBlocksRetrieved * pinstance->num_channel;		

    pos = pinstance->current_pos + nBlocksRetrieved;
    pinstance->current_pos = pos;
    
    return nread;
}

static spBool spSeekPluginMonkey(void *instance, spLong pos)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;
    
    if (pinstance->pAPEDecompress == NULL) return SP_FALSE;
    
    if (pinstance->pAPEDecompress->Seek(pos) != 0) { 
	return SP_FALSE;
    }
    pinstance->current_pos = pos;
    
    return SP_TRUE;
}

static spLong spGetPluginTotalLengthMonkey(void *instance)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;
    
    return pinstance->total_length;
}
